#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Decomposing inshore time series as QuadTrendWithBrk+LinTrendSeasonWithBrk+Residuals
Annual averaging of the residuals to focus on longer time scales
Estimating covariance matrix between the residuals at the 226 glaciers

@author: vincent
"""

import os
import sys
import copy
import csv
import numpy as np
import matplotlib.pyplot as plt
import pyproj
from scipy.linalg import cholesky
from sklearn.covariance import GraphicalLassoCV
from matplotlib import cm
from matplotlib.colors import Normalize 

from codeFunctions import getDevFromYearlyMean
from codeFunctions import piecewisePolyFitSingleBreak
from codeFunctions import fromMonthlyToAnnual


'''
The sparse correlation matrix can be used to generate noise terms
with the desired correlation properties between glaciers.
If the series is represented as an ARMA process, use the calibrated
noise variances of the ARMA models to convert the correlation matrix 
to a covariance matrix (not the variances of the time series themselves)
Noise variance of an ARMA process < Variance of the time series
'''

saveCorrMat        = True
saveCorrMatMembers = False
cwd                = os.getcwd()+'/'

SelModel      = 'MIROCES2L'
SelExperiment = 'hist2100ssp585'
SelCriterion  = 'BIC' #'SigParam'/'BIC'
FitCoef       = 'Averaged' #'BestMember'/'Averaged'

plotmembers   = []
plotAverage   = False

DepthRange       = [0,500]
ShallowThreshold = 100
QMperiodObs      = [1950,2015]

### Parameters ###
SsnDateBreak  = 2015
TrendNbBreaks = 1
TrendOrderFit = 2
SsnOrderFit   = 1

transformer1 = pyproj.Transformer.from_crs('epsg:4326','epsg:3413') #transformer of lat-lon to x-y(3413)
transformer2 = pyproj.Transformer.from_crs('epsg:3413','epsg:4326') #transformer of x-y(3413) to lat-lon


GlFrontloc          = cwd+'InputOutput/frontGlacierPos.csv'
InshModTFloc        = cwd+f'InputOutput/inshoreTFseries{SelModel}_{SelExperiment}_QMen4anlnrnbTFdpavg_Dp{DepthRange[0]}to{DepthRange[1]}_bathymin{ShallowThreshold}_ObsPer{QMperiodObs[0]}to{QMperiodObs[1]}.csv'
infoInshModloc      = cwd+f'InputOutput/infoIDinshoreTFseries{SelModel}_{SelExperiment}_QMen4anlnrnbTFdpavg_Dp{DepthRange[0]}to{DepthRange[1]}_bathymin{ShallowThreshold}_ObsPer{QMperiodObs[0]}to{QMperiodObs[1]}.csv'
LocFittedARMAparams = cwd+f'InputOutput/fitARMAparams_allglaciersinshore_{SelModel}_{SelExperiment}_criterion{SelCriterion}_fitcoef{FitCoef}.csv'
DirOutputs          = cwd+'InputOutput/'

### Load glacier front positions ###
dataGl    = np.genfromtxt(GlFrontloc,dtype=str,delimiter=',')
nGl       = np.shape(dataGl)[0]
nameGl0    = dataGl[:,0]
xFrGl0     = dataGl[:,1].astype(float)
yFrGl0     = dataGl[:,2].astype(float)
# Sorting glaciers in clock-wise order #
ggymax     = np.argmax(yFrGl0)
glymax     = np.max(yFrGl0)
glxdivide  = xFrGl0[ggymax]
indsE      = xFrGl0-glxdivide>0
indsW      = xFrGl0-glxdivide<0
yposE      = np.copy(yFrGl0[indsE])
yposW      = np.copy(yFrGl0[indsW])
srtE       = np.sort(yposE)
srtW       = np.sort(yposW)
indsSort   = [ggymax]
for yvalue in np.flip(srtE):
    jj = np.where(yFrGl0==yvalue)[0][0]
    indsSort.append(jj)
for yvalue in srtW:
    jj = np.where(yFrGl0==yvalue)[0][0]
    indsSort.append(jj)
indsSort = np.array(indsSort)
nameGl   = nameGl0[indsSort]
xFrGl    = xFrGl0[indsSort]
yFrGl    = yFrGl0[indsSort]

latFrGl   = np.zeros(nGl)
lonFrGl   = np.zeros(nGl)
for ll in range(nGl):
    (lat0,lon0) = transformer2.transform(xFrGl[ll],yFrGl[ll])
    latFrGl[ll] = np.copy(lat0)
    lonFrGl[ll] = np.copy(lon0)

### Load reconstructed inshore Model TF time series ###
dataTFinshMod    = np.genfromtxt(InshModTFloc,dtype=float,delimiter=',')
timing           = dataTFinshMod[0,:]
ntime            = len(timing)
stepsperyr       = 12
srTFinshMod0     = dataTFinshMod[1:,:]
datainfoinshMod  = np.genfromtxt(infoInshModloc,dtype=str,delimiter=',')
memberInshore    = datainfoinshMod[:,1]
uniqueMembers    = np.unique(memberInshore)
glacierInshore0  = datainfoinshMod[:,2]
nGl              = len(nameGl)
nMb              = len(uniqueMembers)
        
for mm,member in enumerate(uniqueMembers):
    mmrows           = memberInshore==member
    mmsrTFinshMod0   = srTFinshMod0[mmrows,:]
    mmglacierinsh0   = glacierInshore0[mmrows]
    isortInsh        = []
    for ii,myname in enumerate(nameGl):
        ind       = np.where(mmglacierinsh0==myname)[0][0]
        isortInsh.append(ind)
    mmglaciers   = mmglacierinsh0[isortInsh]
    mmseries0    = mmsrTFinshMod0[isortInsh]
    
    ### Process each glacier time series
    for gg,srorg in enumerate(mmseries0):
        # Remove trend with breakpoint and 2nd order polynomial #
        if(TrendNbBreaks==1 and TrendOrderFit==2):
            params,DateBreak,preBrk,postBrk = piecewisePolyFitSingleBreak(timing,srorg,Order=2)
            b2pre,b1pre,b0pre,b2post,b1post,b0post = params
            tmpre   = timing[preBrk]-timing[0]
            tmpost  = timing[postBrk]-DateBreak
        trendcomp = -1*np.mean(srorg)+np.append(b0pre+b1pre*tmpre+b2pre*tmpre**2,
                                                b0post+b1post*tmpost+b2post*tmpost**2)
        srdtr     = srorg-trendcomp
        mnsrdtr   = np.mean(srdtr)
        
        # Remove seasonality #
        fullmonthlydevs    = getDevFromYearlyMean(timing,srdtr)
        srSsn     = np.zeros(len(timing))
        if(SsnOrderFit==1):
            for mth in range(12):
                # Find all indices of month mth #
                imth          = np.zeros(12).astype(bool)
                imth[mth]     = True
                imth          = np.tile(imth,len(timing/12)+1)[0:len(timing)]
                params,brk,iipre,iipost   = piecewisePolyFitSingleBreak(timing[imth],fullmonthlydevs[imth],Order=1,BreakPoint=SsnDateBreak)
                b1pre,b0pre,b1post,b0post = params
                # Reconstruct seasonality #
                srSsn[imth]  = np.append(b0pre+b1pre*(timing[imth][iipre]-timing[imth][0]),
                                         b0post+b1post*(timing[imth][iipost]-brk))

        # Save residuals #
        monthlyrsd   = srdtr-mnsrdtr-srSsn 
        timingyr,yearlyrsd    = fromMonthlyToAnnual(timing,monthlyrsd)
        if(gg==0):
            mmrsd             = np.copy(yearlyrsd)
        else:
            mmrsd             = np.row_stack((mmrsd,yearlyrsd))

    ### Empirical covariance matrix ###
    empCov   = np.cov(mmrsd) #empirical covariance matrix
    empCorr  = np.corrcoef(mmrsd) #empirical correlation matrix
    empSdevs = np.diag(np.sqrt(np.diag(empCov))) #diagonal matrix of std devs
        
    ### Generate random noise from empirical correlation ###
    sampleNoise = np.random.multivariate_normal(mean=np.zeros(nGl),cov=empCorr,size=ntime)
    
    ### Fit Sparse correlation matrix with graphical lasso ###
    glMod  = GraphicalLassoCV()
    glMod.fit(sampleNoise)
    glCorr = glMod.covariance_
    
    ### Save for averaging ###
    if(mm==0):
        sumEmpCorr = np.copy(empCorr)
        sumSpCorr  = np.copy(glCorr)
    elif(mm>0):
        sumEmpCorr += empCorr
        sumSpCorr  += glCorr

    ### Generate random noise terms from the graphical lasso correlation ###
    cholglCorr          = cholesky(glCorr,lower=True)
    noiseStandardNormal = np.random.normal(loc=0,scale=1,size=(nGl,ntime))
    generatedNoise      = empSdevs @ cholglCorr @ noiseStandardNormal
    
    ### Saving ###
    if(saveCorrMatMembers):
        nameout = f'correlationMatrix_allglaciersinshore_{SelModel}_M{member}_{SelExperiment}.csv'
        locout  = DirOutputs+nameout
        with open(locout,'w',newline='') as outputcsv:
            writer = csv.writer(outputcsv)
            writer.writerow(['']+nameGl.tolist())
            for ii in range(nGl):
                writer.writerow([nameGl[ii]]+glCorr[ii,:].tolist())
    
    if(member in plotmembers):
        ### Plotting ###
        valmin    = 0
        valmax    = +1
        myticks   = np.arange(0,nGl,step=50)
        cbarticks = np.round(np.arange(valmin,valmax+0.01,step=0.2),decimals=1)
        mycmap    = plt.cm.get_cmap('YlGn')
        if(member in plotmembers):
            fig = plt.figure(figsize=[12,6])
            ax  = plt.subplot(121)
            ax.set_position([0.04,0.03,0.4,0.94])
            ax.imshow(empCorr,interpolation='nearest',vmin=valmin,vmax=valmax,cmap=mycmap)
            ax.set_title('Empirical correlation matrix',fontsize=14)
            ax.set_xticks(myticks),ax.set_yticks(myticks)
            ax.set_xticklabels(myticks,fontsize=13),ax.set_yticklabels(myticks,fontsize=13)
            ax  = plt.subplot(122)
            ax.set_position([0.49,0.03,0.4,0.94])
            ax.imshow(glCorr,interpolation='nearest',vmin=valmin,vmax=valmax,cmap=mycmap)
            ax.set_title('Sparse correlation matrix',fontsize=14)
            ax.set_xticks(myticks),ax.set_yticks(myticks)
            ax.set_xticklabels(myticks,fontsize=13),ax.set_yticklabels(myticks,fontsize=13)
            cax  = fig.add_axes([0.91,0.2,0.01,0.6])
            cbar = plt.colorbar(cm.ScalarMappable(norm=Normalize(vmin=valmin,vmax=valmax),cmap=mycmap),cax=cax)
            cbar.set_ticks(cbarticks)
            cbar.set_ticklabels(cbarticks,fontsize=13)
            cbar.set_label('Correlation',fontsize=14,rotation=-90,labelpad=30)
            
            if(False):
                fig = plt.figure(figsize=[12,6])
                ax  = plt.subplot(111)
                ax.scatter(np.arange(0,nGl),np.sqrt(sig2noiseStat))
                fig.tight_layout()

avgEmpCorr = sumEmpCorr/nMb
avgSpCorr  = sumSpCorr/nMb
### Saving ###
if(saveCorrMat):
    nameout = f'correlationMatrix_allglaciersinshore_{SelModel}_MembersAverage_{SelExperiment}.csv'
    locout  = DirOutputs+nameout
    with open(locout,'w',newline='') as outputcsv:
        writer = csv.writer(outputcsv)
        writer.writerow(['']+nameGl.tolist())
        for ii in range(nGl):
            writer.writerow([nameGl[ii]]+avgSpCorr[ii,:].tolist())

if(plotAverage):
    ### Plotting ###
    valmin    = 0
    valmax    = +1
    myticks   = np.arange(0,nGl,step=50)
    cbarticks = np.round(np.arange(valmin,valmax+0.01,step=0.2),decimals=1)
    mycmap    = plt.cm.get_cmap('YlGn')
    fig = plt.figure(figsize=[12,6])
    ax  = plt.subplot(121)
    ax.set_position([0.04,0.03,0.4,0.94])
    ax.imshow(avgEmpCorr,interpolation='nearest',vmin=valmin,vmax=valmax,cmap=mycmap)
    ax.set_title('Empirical correlation matrix',fontsize=14)
    ax.set_xticks(myticks),ax.set_yticks(myticks)
    ax.set_xticklabels(myticks,fontsize=13),ax.set_yticklabels(myticks,fontsize=13)
    ax.text(0,-5,'(a)',fontsize=13)
    ax  = plt.subplot(122)
    ax.set_position([0.49,0.03,0.4,0.94])
    ax.imshow(avgSpCorr,interpolation='nearest',vmin=valmin,vmax=valmax,cmap=mycmap)
    ax.set_title('Sparse correlation matrix',fontsize=14)
    ax.set_xticks(myticks),ax.set_yticks(myticks)
    ax.set_xticklabels(myticks,fontsize=13),ax.set_yticklabels(myticks,fontsize=13)
    ax.text(0,-5,'(b)',fontsize=13)
    cax  = fig.add_axes([0.91,0.2,0.01,0.6])
    cbar = plt.colorbar(cm.ScalarMappable(norm=Normalize(vmin=valmin,vmax=valmax),cmap=mycmap),cax=cax)
    cbar.set_ticks(cbarticks)
    cbar.set_ticklabels(cbarticks,fontsize=13)
    cbar.set_label('Correlation',fontsize=14,rotation=-90,labelpad=30)
    

